'
'	pb_sql3.bas
'
'	DESCRIPTION
'	wrapper for sqlite3.dll. sqlite3.dll is the public domain sqlite3 sql
'	database engine from http://www.sqlite.org Please see that site for
'	availability of the sqlite3.dll file.
'
'	PUBLIC DOMAIN
'	Placed in the public domain by Don Dickinson,
'	http://www.greatwebdivide.com
'	mailto:don@greatwebdivide.com
'
'	LICENSE
'	This code is free for all to use. No guarantees or warranties of any
'	kind are granted. You are using this free code at your own risk. Its
'	use implies that you accept this fact. This file is dependent on the
'	following source files made by Don Dickinson.
'		pb_sqll3.inc 	-> inclean'd winapi declarations
'		pb_mem.bas 		-> memory handling
'		dd_link.bas 	-> linked lists
'
'	The linked list library is extensive. Only a few of the calls in it
'	are actually used. It would be possible to make the compilation smaller
'	by eliminating unused calls from it. If you do this, please rename the
'	file so it doesn't conflict with anyone using the original file.
'
'	COMPILING
'	This requires pbdll 6.1 or greater to compile as a dll.
'	It may also be compiled into your code with pbcc2.1 or greater.
'	There is a small demo program in this header. Just un-comment the
'	%COMPILE_TEST_PROGRAM line below and compile with pbcc2.1 or greater.
'	To compile as a dll un-rem the %COMPILE_DLL equate below.
'
'	CHANGES FROM SQLITE v2 CODE
'		>> Major change is that the error message variable
'			parameter is not passed as a parameter.
'			if a call fails use the sql3errCode() and/or sql3errMsg(O)
'			functions to obtain the error information. The exception
'			to this is the sql3useDB command which keeps the error
'			parameter in it's call definition.
'		>> Row ID's are now Quads
'	   >> MAjor change to sql3nextRow: this function now
'			returns a constant instead of %true or %false
'
'	HISTORY
'		8-4-2004 DD
'		>> Original translation from the 2.x api
'
#if not %def(%PB_SQL3_BAS)
%PB_SQL3_BAS = 1

'- NOTE: if you uncomment the line below
'			and compile with pbcc2.1 or greater
'			you will be compiling a small test
'			program that creates a database table
'			and puts some data into it. Uncomment
'			the %COMPILE_DLL flag to compile as
'			a dll with pbdll6.1 or greater.
'
%COMPILE_TEST_PROGRAM = 1
'%COMPILE_DLL = 1

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FOR TEST PROGRAM ONLY (see %COMPILE_TEST_PROGRAM equate above)
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#if %def(%COMPILE_TEST_PROGRAM)
   #compile exe
   #dim all
   #include "pb_sql3.inc"			'inclean'd api

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  COMPILE AS A DLL (see %COMPILE_DLL equate above)
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#elseif %def(%COMPILE_DLL)
   #compile dll "c:\winnt\system32\pbsqlite3.dll"
   #dim all
   #include "pb_sql3.inc"			'inclean'd api

#endif
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#include "pb_mem.bas"			'memory allocation
#include "dd_link.bas"			'linked lists

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FOR TEST PROGRAM ONLY
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#if %def(%COMPILE_TEST_PROGRAM)
#include "dd_stdout.bas"
#endif
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	GLOBALS
'	needed to keep track of options and initialization status.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
global g_sql3_Init as long
global g_sql3_BusyRetry as long
global g_sql3_RetryWait as long

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	CONSTANTS
'	first ones are defaults for the g_sql3* variables
'	the rest are return codes for the various sqlite functions
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
%PBSQL3_DEFAULT_BUSY_RETRY = 10
%PBSQL3_DEFAULT_RETRY_WAIT = 500

%PBSQL3_DATA = 0 	'Data is available for reading with sql3field
%PBSQL3_EOF  = 1  'No more data to read
%PBSQL3_BUSY = 2  'Timeout, system is busy
%PBSQL3_ERR  = 3  'Some sort of error occurred


%SQLITE3_ISO8859         =   1
%SQLITE3_OK              =   0   '*/ Successful result */
%SQLITE3_ERROR           =   1   '*/ SQL error or missing database */
%SQLITE3_INTERNAL        =   2   '*/ An internal logic error in SQLite */
%SQLITE3_PERM            =   3   '*/ Access permission denied */
%SQLITE3_ABORT           =   4   '*/ Callback routine requested an abort */
%SQLITE3_BUSY            =   5   '*/ The database file is locked */
%SQLITE3_LOCKED          =   6   '*/ A table in the database is locked */
%SQLITE3_NOMEM           =   7   '*/ A malloc() failed */
%SQLITE3_READONLY        =   8   '*/ Attempt to write a readonly database */
%SQLITE3_INTERRUPT       =   9   '*/ Operation terminated by sqlite3_interrupt() */
%SQLITE3_IOERR           =   10   '*/ Some kind of disk I/O error occurred */
%SQLITE3_CORRUPT         =   11   '*/ The database disk image is malformed */
%SQLITE3_NOTFOUND        =   12   '*/ (Internal Only) Table or record not found */
%SQLITE3_FULL            =   13   '*/ Insertion failed because database is full */
%SQLITE3_CANTOPEN        =   14   '*/ Unable to open the database file */
%SQLITE3_PROTOCOL        =   15   '*/ Database lock protocol error */
%SQLITE3_EMPTY           =   16   '*/ (Internal Only) Database table is empty */
%SQLITE3_SCHEMA          =   17   '*/ The database schema changed */
%SQLITE3_TOOBIG          =   18   '*/ Too much data for one row of a table */
%SQLITE3_CONSTRAINT      =   19   '*/ Abort due to contraint violation */
%SQLITE3_MISMATCH        =   20   '*/ Data type mismatch */
%SQLITE3_MISUSE          =   21   '*/ Library used incorrectly */
%SQLITE3_NOLFS           =   22   '*/ Uses OS features not supported on host */
%SQLITE3_AUTH            =   23   '*/ Authorization denied */
%SQLITE3_ROW             =   100  '*/ sqlite3_step() has another row ready */
%SQLITE3_DONE            =   101  '*/ sqlite3_step() has finished executing */


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	STRUCTURES
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	This structure is used by this wrapper only. it is not part of sqlite.
'	A pointer to this structure is passed as the pExecute parameter in
'	various wrapper functions.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
type INTERNAL_VMM_TYPE
	pVMM as long
	columnCount as long
	columnData as long
end type

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	FUNCTION DECLARATIONS
'	these declare the sqlite3 function calls in sqlite3.dll
'	note that they are all CDECL. if you do it without this, errors and
'	strange things will happen in your program eventually.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
declare function sqlite3_open cdecl lib "sqlite3.dll" alias "sqlite3_open" _
		( zFilename as asciiz, hSQLite as long ) as long
declare sub sqlite3_close cdecl lib "sqlite3.dll" alias "sqlite3_close" _
		( byval hSQLite as long )

declare function sqlite3_last_insert_rowid cdecl lib "sqlite3.dll" alias "sqlite3_last_insert_rowid" _
		( byval hSQLite as long ) as quad

declare function sqlite3_changes cdecl lib "sqlite3.dll" alias "sqlite3_changes" _
		( byval hSQLite as long ) as long
declare sub sqlite3_interrupt cdecl lib "sqlite3.dll" alias "sqlite3_interrupt" _
		( byval hSQLite as long )
declare function sqlite3_complete cdecl lib "sqlite3.dll" alias "sqlite3_complete" _
		( zSQL as asciiz ) as long
declare function sqlite3_get_table cdecl lib "sqlite3.dll" alias "sqlite3_get_table" _
		( byval hSQLite as long, zSQL as asciiz, lpTable as long, iRow as long, _
		  iCol as long, lpErrMsg as long ) as long
declare function sqlite3_free_table cdecl lib "sqlite3.dll" alias "sqlite3_free_table" _
		( byval lpTable as long ) as long

declare function sqlite3_prepare cdecl lib "sqlite3.dll" alias "sqlite3_prepare" _
		( byval hSQLite as long, zSQL as asciiz, byval iLen as long, _
		  ppVMM as long ptr, pzTail as asciiz ptr ) as long
declare function sqlite3_finalize cdecl lib "sqlite3.dll" alias "sqlite3_finalize" _
		( byval pStmt as long ) as long
declare function sqlite3_step cdecl lib "sqlite3.dll" alias "sqlite3_step" _
		( byval pStmt as long ) as long
declare function sqlite3_errcode cdecl lib "sqlite3.dll" alias "sqlite3_errcode" _
		( byval hSQLite as long ) as long
declare function sqlite3_errmsg cdecl lib "sqlite3.dll" alias "sqlite3_errmsg" _
		( byval hSQLite as long ) as long

declare function sqlite3_create_function cdecl lib "sqlite3.dll" alias "sqlite3_create_function" _
		( byval hSQLite as long, zName as asciiz, byval nArg as long, _
		  byval eTextRep as long, byval pAny as any, byval pFunction as long, _
		  byval pStep as long, byval pFinal as long ) as long
declare function sqlite3_create_collation cdecl lib "sqlite3.dll" alias "sqlite3_create_collation" _
		( byval hSQLite as long, zName as asciiz, byval pref16 as long, _
		  byval pAny as long, byval pCompare as long) as long

declare function sqlite3_column_count cdecl lib "sqlite3.dll" alias "sqlite3_column_count" _
		( byval hStmt as long ) as long
declare function sqlite3_column_name cdecl lib "sqlite3.dll" alias "sqlite3_column_name" _
		( byval hStmt as long, byval iCol as long ) as long
declare function sqlite3_column_text cdecl lib "sqlite3.dll" alias "sqlite3_column_text" _
		( byval hStmt as long, byval iCol as long ) as long

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3errCode
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3errCode alias "sql3errCode" _
		( byval hSQLite as long ) export as long

	function = sqlite3_errcode(hSQLite)

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3errMsg
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3errMsg alias "sql3errMsg" _
		( byval hSQLite as long ) export as string

	dim pzErr as asciiz ptr

	pzErr = sqlite3_errmsg(hSQLite)
	if pzErr = 0 then
		function = ""
	else
		function = @pzErr
	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3FreeMem
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sql3FreeMem alias "sql3FreeMem" ( byval pMem as long ) export

	if pMem then freeMem pMem

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3Init
'
'	initializes the sqlite wrapper. call this exactly once at the beginning
'	of your app before other calls. Call the sister function sqliteDone
'	exactly once before finishing.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sql3Init alias "sql3Init" export

	if g_sql3_Init then exit sub

	g_sql3_Init = %true
	g_sql3_BusyRetry = %PBSQL3_DEFAULT_BUSY_RETRY
	g_sql3_RetryWait = %PBSQL3_DEFAULT_RETRY_WAIT

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3Done
'
'	Cleans up the sqlite wrapper. call this exactly once before your
'	app exits.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sql3Done alias "sql3Done" () export

	if g_sql3_Init = %false then exit sub
	g_sql3_Init = %false

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  sql3LastID
'  returns the primary integer key of the last row inserted
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3LastID alias "sqliteLastID" _
      ( byval hDB as long ) export as quad

   function = sqlite3_last_insert_rowid(hDB)

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3UseDB
'
'	Opens a database file or creates one if it doesn't already exist.
'  It always opens in read-write mode unless the file itself is readonly.
'  Then it opens in Read-Only mode. this is the current behavior of the
'  sqlite engine. the sqlite3_open "mode"parameter is ignored.
'
'	Returns:
'		0 on failure and fills sErr with a text error message.
'		a non-zero handle (hDB parameters) to be passed to other engine calls.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3UseDB alias "sql3UseDB" _
		( sDBName as string, sErr as string ) export as long

	dim hDB as long
	dim iResult as long
	dim pzError as asciiz ptr

	if g_sql3_Init = %false then sql3Init

	iResult = sqlite3_open ( sDBName + $nul, hDB  )
	if iResult = %SQLITE3_OK then
		function = hDB
	else
		if hDB = 0 then
			sErr = "Unhandled internal error in sql3useDB. sqlite3_open " + _
					"did not return an sqlite handle."
		else
			sErr = format$(sql3errCode(hDB)) + ": " + sql3errMsg(hDB)
			sqlite3_close hDB
		end if
		function = 0
	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3CloseDB
'
'	Closes an sqlite database that was opened with sqliteUseDB
'	Pass it the hDB handle returned by the sqliteUseDB call.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
sub sql3CloseDB alias "sqliteCloseDB" ( hDB as long ) export

	if g_sql3_Init = %false then sql3Init
	sqlite3_close hDB

end sub

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3Execute
'
'	Executes an sql statement. Use its return value in the sqliteNextRow
'	and sqliteExecuteDone functions.
'
'	Returns:
'		0 		on failure and fills the sErr parameter with a description
'				of the error.
'		<> 0  on success. This is a pExecute handle to be passed to
'				subsequent calls to sqliteNextRow and sqliteExecuteDone
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3Execute alias "sql3Execute" _
		( hDB as long, sSQL as string ) export as long

	dim bOk as long
	dim i as long
	dim iResult as long
	dim pzTail as asciiz ptr
	dim ppVmm as long ptr
	dim iFoo as long
	dim colName as string
	dim pzColName as asciiz ptr
	dim pResult as INTERNAL_VMM_TYPE ptr

	if g_sql3_Init = %false then sql3Init

	bOk = %true

	iFoo = 0
	ppVmm = varptr(iFoo)

	'- get the vm for processing sql statement
	iResult = sqlite3_prepare(hDB, sSQL + $nul, -1, ppVmm, pzTail)
	if iResult <> %SQLITE3_OK then
		bOk = %false
		goto sql3Execute_done
	end if

sql3Execute_done:

	'- error
	if bOk = %false then
		if @ppVmm <> 0 then sqlite3_finalize @ppVmm
		function = 0

	'- ok
	else
		pResult = getMem(sizeof(@pResult))
		@pResult.pVMM = @ppVmm
		@pResult.columnCount = sqlite3_column_count(@pResult.pVMM)
		@pResult.columnData = llAdd(0, "**root**", "**root**")

		'- get information about the columns
		for i = 1 to @pResult.columnCount
			pzColName = sqlite3_column_name(@pResult.pVMM, i-1)
			colName = trim$(lcase$(@pzColName))
			llAdd @pResult.columnData, colName, ""
		next i

		function = pResult

	end if

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3ExecuteDone
'
'	Cleans up after a successful sqliteExecute command.
'	pass the return value of sqliteExecute* in the pExecute parameter
'	sErr contains the error message if this generates an error.
'
'	Returns:
'		%true 		on success and fills the sErr parameter
'		%false 		on failure
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3ExecuteDone alias "sql3ExecuteDone" _
		( byval pExecute as long ) export as long

	dim bOk as long
	dim iResult as long
	dim pInt as INTERNAL_VMM_TYPE PTR

	if g_sql3_Init = %false then sql3init

	pInt = pExecute
	bOk = %true

	'- huh? this shouldn't happen
	if pInt = 0 then
		function = bOk
		exit function
	end if

	'- free the virtual machine.
	if @pInt.pVMM then sqlite3_finalize @pInt.pVMM

	'- free the field pointers
	if @pInt.columnData then llFree @pInt.columnData

	'- free the structure
	freeMem pInt

	function = bOk

end function


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3NextRow
'
'	Read the next row of data. You should call this function until
'	it returns %false (indicating failure or end of data)
'
'	Returns
'		%PBSQL3_DATA Data is available for reading with sql3field
'		%PBSQL3_EOF  No more data to read
'		%PBSQL3_BUSY Timeout, system is busy
'		%PBSQL3_ERR  Some sort of error occurred
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3NextRow alias "sql3NextRow" _
		( byval pExecute as long ) export as long

	dim i as long
	dim iCount as long
	dim iCols as long
	dim iResult as long
	dim pField as long
	dim pzValue as asciiz ptr
	dim pInt as INTERNAL_VMM_TYPE PTR

	if g_sql3_Init = %false then sql3init
	pInt = pExecute

	iResult = sqlite3_step(@pInt.pVMM)
	do
		if iResult = %SQLITE3_BUSY then
			incr iCount
			if iCount > g_sql3_BusyRetry then
				function = %PBSQL3_DATA
				exit do
			end if
			apiSleep g_sql3_RetryWait

		elseif iResult = %SQLITE3_DONE then
			function = %PBSQL3_EOF
			exit do

		elseif iResult = %SQLITE3_ROW then

			'- read the data into the linked list
			pField = @pInt.columnData
			for i = 1 to @pInt.columnCount
				pzValue = sqlite3_column_text(@pInt.pVMM, i-1)
				pField = llNext(pField)
				llUpdate pField, @pzValue
			next i

			function = %PBSQL3_DATA
			exit do

		else				'%SQLITE3_ERROR or %SQLLIST_MISUSE
			function = %PBSQL3_ERR
			exit do

		end if
	loop

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3Field
'
'	Returns the string contents of a field in the row obtained by the
'	last call to sqliteNextRow
'
'	Returns:
'		the string contents of the field
'		an empty string if the field name is invalid.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3Field alias "sql3Field" _
		( byval pExecute as long, fieldName as string ) export as string

	dim pInt as INTERNAL_VMM_TYPE PTR

	pInt = pExecute
	if pInt = 0 then
		function = ""
		exit function
	end if

	fieldName = trim$(lcase$(fieldName))
	function = llDataByName(@pInt.columnData, fieldName)

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	sql3ExecuteNoReturn
'
'	Executes a SQL statement and just determines if it was successful. It
'	does not report an row data. Use this for INSERT, UPDATE, DELETE statements
'	where you won't expect a result set. It cleans up after itself, so there
'	is no need to call sqlliteExecuteDone
'
'	Returns
'		%true		on success
'		%false	on failure
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3ExecuteNoReturn alias "sql3ExecuteNoReturn" _
		( byval hDB as long, sSQL as string ) export as long

	dim iResult as long
	dim hSQL as long
	dim sFoo as string

	hSQL = sql3Execute(hDB, sSQL)
	if hSQL = 0 then
		function = %false
		exit function
	end if

	iResult = sql3NextRow(hSQL)
	do until iResult > 0
		iResult = sql3NextRow(hSQL)
	loop

	sql3ExecuteDone hSQL

	if iResult = %PBSQL3_EOF then
		function = %true
	else
		function = %false
	end if


end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  sqliteClean
'  cleans a string for the insert statement
'  VB/PB style (escapes single quotes)
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3Clean alias "sql3Clean" _
      ( sIncoming as string ) export as string

   dim sReturn as string

   '- don't change the original
   sReturn = sIncoming
   replace "'" with "''" in sReturn
   function = sReturn

end function

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'  sql3TableExists
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function sql3TableExists alias "sql3TableExists" _
      ( byval hDB as long, sTableName as string ) export as long

	dim iResult as long
   dim hSQL as long
   dim sSQL as string

   function = %false

   sSQL = "SELECT name FROM sqlite_master WHERE (type='table') AND " + _
         "name='" + sql3Clean(sTableName) + "'"

   hSQL = sql3Execute(hDB, sSQL)
   if hSQL = 0 then
      exit function
   end if

	iResult = sql3NextRow(hSQL)
   do until iResult <> 0
      if trim$(lcase$(sql3Field(hSQL, "tbl_name"))) = _
            trim$(lcase$(sTableName)) then
         function = %true
         exit do
      end if
		iResult = sql3NextRow(hSQL)
   loop

end function

#if %def(%COMPILE_TEST_PROGRAM)
$TEST_FILE = ".\test_demo_db.sqlite"
$SQL_TABLE_CREATE = "CREATE TABLE t_test (nm char(40), count_int int)"

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'	pbmain
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
function pbmain

	dim hDB as long
	dim hSQL as long
	dim i as long
	dim iResult as long
	dim iCount as long
	dim startTime as double
	dim sSQL as string
	dim sErr as string

	'- Always the first call your app should make.
	sql3init

	'- Delete the test database if it exists. Then Open it.
	if dir$($TEST_FILE) <> "" then
		kill $TEST_FILE
	end if

	hDB = sql3UseDB($TEST_FILE, sErr)
	if hDB = 0 then
		stdout "error: " + sErr
		goto pbmain_done
	end if

	'- Execute the sql statement to create a database
	stdout "OK about to execute"
	if sql3ExecuteNoReturn(hDB, $SQL_TABLE_CREATE) = %false then
		stdout "ERROR creating table"
		stdout sErr
		goto pbmain_done
	end if
	stdout "Table created."

	'- Begin a transaction. If you don't wrap your multiple transactions
	'	inside of a transaction, they will go much, much slower.
	'
	if sql3ExecuteNoReturn(hDB, "BEGIN TRANSACTION") = %false then
		stdout "ERROR beginning transaction"
		stdout sErr
		goto pbmain_done
	end if

	startTime = timer

	'- Insert 100 records
	for i = 1 to 100
		if i mod 10 = 0 then
			stdoutCL "Inserted " + format$(i)
		end if
		sSQL = "INSERT INTO t_test (nm,count_int) " + _
				"VALUES ('Name " + format$(i) + "', " + format$(i) + ")"
		if sql3ExecuteNoReturn(hDB, sSQL) = %false then
			stdout "ERROR inserting record: " + sql3ErrMsg(hDB)
			goto pbmain_done
		end if

	next i

	'- Commit them.
	stdout "END TRANSACTION"
	if sql3ExecuteNoReturn(hDB, "END TRANSACTION") = %false then
		stdout "ERROR ending transaction" + sql3ErrMsg(hDB)
		stdout sErr
		goto pbmain_done
	end if

	stdout
	stdout "time = " + format$(timer - startTime)
	stdout "Done inserting"

   	'- After we finish inserting, then read back the count.
	'	notice i loop through the records. this isn't probably
	'	necessary because only one record will be returned. That
	'	record will have only one field: the count.
	'
	sSQL = "SELECT count(*) AS icount FROM t_test"
	hSQL = sql3Execute(hDB, sSQL)

	if hSQL = 0 then
		stdout "error selecting count: " + sql3errMsg(hDB)
	else
		do until sql3NextRow(hSQL)
			stdout "count = " + sql3Field(hSQL, "icount")
		loop
	end if

	'- After we're done looping, we have to cleanup the
	'	execution of the sql or we get a memory leek.
	'
	if sql3ExecuteDone(hSQL) = %false then
		stdout "ERROR finishing sql: " + sql3errMsg(hDB)
	end if

	stdout "Press any key to read back the records (esc to quit)"
	if waitkey$ = $esc then goto pbmain_done

	'- Select all of the records
	sSQL = "SELECT nm, count_int, nm AS nml FROM t_test ORDER BY nm"
	hSQL = sql3Execute(hDB, sSQL)
	if hSQL = 0 then
		stdout "Error executing SQL"
		stdout sql3errMsg(hDB)
		stdout "SQL:"
		stdout sSQL
		goto pbmain_done
	end if

	'- loop through them
	cls
	iCount = 0
	iResult = sql3NextRow(hSQL)
	do until iResult
		incr iCount
		if iCount mod 23 = 0 then
			stdout "Press any key to continue"
			waitkey$
			cls
		end if
		stdout sql3Field(hSQL, "count_int") + ": " + sql3Field(hSQL, "nml")
		iResult = sql3NextRow(hSQL)
	loop

	if iResult <> %PBSQL3_EOF then
		stdout "Error occurred reading records " + sql3errMsg(hDB)
	end if

	'- cleanup
	if sql3ExecuteDone(hSQL) = %false then
		stdout "Error closing sql statement: " + sql3errMsg(hDB)
	end if

pbmain_done:

	if hDB then
		stdout "Closing database"
		sql3CloseDB hDB
	end if

	sql3Done

	stdout "DONE"
	waitkey$

end function
#endif
#endif
